#include "model.hxx"

// For `std::logic_error` and `std::invalid_argument`:
#include <stdexcept>

Player other_player(Player p)
{
    switch (p) {
    case Player::first:
        return Player::second;
    case Player::second:
        return Player::first;
    default:
        throw std::invalid_argument("other_player: not a player");
    }
}

// Constructor for Connect4_model.
//
// The second line (`: board_(m)`) constructs the `board_` member
// variable. Since `board_` is declared like
//
//     std::vector<column_t> board_;
//
// the member initializer is as if we were constructing it like
//
//     std::vector<column_t> board_(m);
//
// which initializes `board_` to have `m` elements, each of which
// is a default-constructed (empty) `column_t`.
Connect4_model::Connect4_model()
        : board_(m)
{ }


const int Connect4_model::k = 4;  // how many to connect
const int Connect4_model::m = 7;  // grid width
const int Connect4_model::n = 6;  // grid height


void Connect4_model::place_token(int col_no)
{
    check_playable_(col_no);
    board_[col_no].push_back(turn_);
    update_winner_and_turn_(col_no);
}

const Connect4_model::column_t& Connect4_model::column(int col_no) const
{
    check_column_(col_no);
    return board_[col_no];
}

bool Connect4_model::is_good_column(int col_no) const
{
    return 0 <= col_no && col_no < m;
}

bool Connect4_model::is_playable(int col_no) const
{
    return turn() != Player::neither &&
           is_good_column(col_no) &&
           column(col_no).size() < Connect4_model::n;
}


void Connect4_model::update_winner_and_turn_(int const col_no)
{
    Posn move{col_no, (int)board_[col_no].size() - 1};

    int below       = count_from_by_(move, 0, -1);
    int left        = count_from_by_(move, -1, 0);
    int right       = count_from_by_(move, 1, 0);
    int above_left  = count_from_by_(move, -1, 1);
    int above_right = count_from_by_(move, 1, 1);
    int below_left  = count_from_by_(move, -1, -1);
    int below_right = count_from_by_(move, 1, -1);

    if (below + 1 >= k ||
        left + 1 + right >= k ||
        above_left + 1 + below_right >= k ||
        above_right + 1 + below_left >= k) {
        winner_ = turn_;
    } else {
        for (column_t const& column : board_) {
            if (column.size() < n) {
                turn_ = other_player(turn_);
                return;
            }
        }
    }

    turn_ = Player::neither;
}

void Connect4_model::check_column_(int col_no) const
{
    if (!is_good_column(col_no))
        throw std::invalid_argument("Model::place_token: column out of bounds");
}

void Connect4_model::check_playable_(int col_no) const
{
    check_column_(col_no);

    if (is_game_over())
        throw std::logic_error("Model::place_token: game over");

    if (!is_playable(col_no))
        throw std::invalid_argument("Model::place_token: column full");
}

int Connect4_model::count_from_by_(Posn start, int dcol, int drow) const
{
    int count = 0;

    for (;;) {
        start.row += drow;
        start.col += dcol;

        if (!is_occupied_(start.col, start.row))
            return count;

        if (board_[start.col][start.row] != turn_)
            return count;

        ++count;
    }
}

bool Connect4_model::is_occupied_(int col_no, int row_no) const
{
    return is_good_column(col_no) &&
           0 <= row_no && row_no < board_[col_no].size();
}

