#include "model.hxx"
#include <catch.hxx>
#include <iostream>
#include <stdexcept>


/////
/////
///// TESTING HELPERS
/////
/////

static const Player RED = Player::first,
                    BLU = Player::second,
                    MT  = Player::neither;

using Model  = Connect4_model;
using Column = Model::column_t;

// Helper struct for checking (writing down, comparing, printing) the
// board state.
struct Board
{
    // Holds the whole board; padded to height `Model::n` with
    // `Player::neither`.
    std::vector<Column> columns;

    // Constructs an empty board.
    Board()
            : columns(Model::m, Column(Model::n, Player::neither))
    { }

    // Constructs a board by copying it from a model.
    Board(const Model& model)
            : Board()
    {
        for (int col_no = 0; col_no < Model::m; ++col_no)
            col(col_no, model.column(col_no));
    }

    // Replaces the column at the given index with the given column.
    Board& col(int col_no, const Column& col)
    {
        columns[col_no] = col;
        columns[col_no].resize(Model::n, Player::neither);
        return *this;
    }
};

bool operator==(const Board& a, const Board& b)
{
    return a.columns == b.columns;
}

// How we'll print players when printing a board.
static char char_of(Player player)
{
    switch (player) {
    case Player::first: return '1';
    case Player::second: return '2';
    case Player::neither: return '_';
    }
}

// Prints a board.
std::ostream& operator<<(std::ostream& os, const Board& board)
{
    os << "board:\n";
    for (int row = Model::n - 1; row >= 0; --row) {
        for (const Column& col : board.columns)
            os << char_of(col[row]);
        os << '\n';
    }
    return os;
}


/////
/////
///// TEST CASES
/////
/////

TEST_CASE("Can construct model")
{
    Model model;
}

TEST_CASE("Game parameters match classic Connect Four")
{
    CHECK( Model::k == 4 );
    CHECK( Model::m == 7 );
    CHECK( Model::n == 6 );
}

TEST_CASE("New model in expected state")
{
    Model c4;

    CHECK(c4 == Board());
    CHECK(c4.turn() == RED);
    CHECK(c4.winner() == MT);
    CHECK_FALSE(c4.is_game_over());
}

TEST_CASE("Player 1 can move")
{
    Model c4;

    c4.place_token(1);

    CHECK(c4 == Board().col(1, {RED}));
    CHECK(c4.turn() == BLU);
    CHECK(c4.winner() == MT);
}

TEST_CASE("Player 2 can play next to player 1")
{
    Model c4;

    c4.place_token(1);
    c4.place_token(2);

    CHECK(c4 == Board().col(1, {RED})
                       .col(2, {BLU}));
}


TEST_CASE("Player 2 can play atop to player 1")
{
    Model c4;

    c4.place_token(1);
    c4.place_token(1);

    CHECK(c4 == Board().col(1, {RED, BLU}));
}

TEST_CASE("Can play a whole game.")
{
    Model c4;

    CHECK(c4.turn() == RED );
    CHECK(c4.winner() == MT );
    c4.place_token(0);
    CHECK(c4 == Board().col(0,{RED}));

    CHECK(c4.turn() == BLU );
    c4.place_token(1);
    CHECK(c4 == Board().col(0, {RED})
                       .col(1, {BLU}));

    CHECK(c4.turn() == RED );
    c4.place_token(1);
    CHECK(c4 == Board().col(0, {RED})
                       .col(1, {BLU, RED}));

    c4.place_token(2);
    CHECK(c4 == Board().col(0, {RED})
                       .col(1, {BLU, RED})
                       .col(2, {BLU}));

    c4.place_token(3);
    CHECK(c4 == Board().col(0, {RED})
                       .col(1, {BLU, RED})
                       .col(2, {BLU})
                       .col(3, {RED}));

    c4.place_token(2);
    CHECK(c4 == Board().col(0, {RED})
                       .col(1, {BLU, RED})
                       .col(2, {BLU, BLU})
                       .col(3, {RED}));

    c4.place_token(2);
    CHECK(c4 == Board().col(0, {RED})
                       .col(1, {BLU, RED})
                       .col(2, {BLU, BLU, RED})
                       .col(3, {RED}));

    c4.place_token(3);
    CHECK(c4 == Board().col(0, {RED})
                       .col(1, {BLU, RED})
                       .col(2, {BLU, BLU, RED})
                       .col(3, {RED, BLU}));

    c4.place_token(2);
    CHECK(c4 == Board().col(0, {RED})
                       .col(1, {BLU, RED})
                       .col(2, {BLU, BLU, RED, RED})
                       .col(3, {RED, BLU}));

    c4.place_token(3);
    CHECK(c4 == Board().col(0, {RED})
                       .col(1, {BLU, RED})
                       .col(2, {BLU, BLU, RED, RED})
                       .col(3, {RED, BLU, BLU}));

    c4.place_token(3);
    CHECK(c4 == Board().col(0, {RED})
                       .col(1, {BLU, RED})
                       .col(2, {BLU, BLU, RED, RED})
                       .col(3, {RED, BLU, BLU, RED}));


    CHECK(c4.winner() == RED );
    CHECK(c4.turn() == MT );

    CHECK_THROWS_AS(c4.place_token(0), std::logic_error);
}

TEST_CASE("full column throws")
{
    Model c4;

    for (int i = 0; i < Model::n; ++i)
        c4.place_token(2);

    CHECK_THROWS_AS(c4.place_token(2), std::invalid_argument);

    c4.place_token(0);

    CHECK(c4 == Board().col(0, {RED})
                       .col(2, {RED, BLU, RED, BLU, RED, BLU}));

    CHECK_THROWS_AS(c4.place_token(2), std::invalid_argument);
}
